// Gray - A simple ray tracing program
// Copyright (C) 2008-2018 Eric Bechet
//
// See the COPYING file for contributions and license information.
// Please report all bugs and problems to <bechet@cadxfem.org>.
//

#include <cmath>
#include <cstdlib>
#include <ostream>
#include <iostream>
#include <iomanip>
#include <vector>
#include "QuasiMonteCarlo.h"
#include "no_omp.h"


#include "framebuffer.h"

void framebuffer::makeimage ( void )
{
  time_t start = time ( 0 );
  srand ( start );
  sc->initCounter();

  int width = getwidth();
  int height = getheight();

  //Clear the buffer
  for ( int i = 0; i < width; i++ )
    for ( int j = 0; j < height; j++ )
      ( *this ) ( i,j ) = Color ( 0,0,0 );

  Camera* camera = sc->getCam();
  Point vectorLook = camera->getDirection() - sc->getViewpoint();
  vectorLook.normalize();
  VdcRandomizer vdcRandomizer;

  int counter = 0;
  #pragma omp parallel
  {
    if ( omp_get_thread_num() == 0 )
    {
      std::cout << "Starting on " << omp_get_num_threads() << " threads. " << std::endl;
      std::cout << "|__________________________________________________|" << std::endl;
      std::cout << "|";
      std::cout.flush();
    }
  }

#ifdef USE_OMP_2_0
  #pragma omp parallel for schedule(dynamic,10)
#else
  #pragma omp parallel for collapse(3) schedule(dynamic,10)
#endif

  for ( int i = 0; i < width; i++ )
  {
    for ( int j = 0; j < height; j++ )
    {
      for ( int k = 0; k < sc->getOversampling(); k++ )
      {
        if ( omp_get_thread_num() == 0 && ( i*50 ) /width >= counter )
        {
          std::cout << "=";
          std::cout.flush();
          counter++;
        }

        float x = i;
        float y = j;
        float dx = k==0?0:vdcRandomizer.getNext ( 3 )-0.5f;
        float dy = k==0?0:vdcRandomizer.getNext ( 5 )-0.5f;

        Ray r;
        get_coords ( x+dx,y+dy,r );
        Ray ray ( r.p, r.v*-1 );
        Color c;
        sc->raytracing ( ray,c );
        c = c * 1.0/sc->getOversampling();
        #pragma omp critical
        {
          ( *this ) ( i,j ) = ( *this ) ( i,j ) + c;
        }
      }
    }
  }

  empty = false;
  std::cout << "|" << std::endl;
  std::cout << "Computed in " << ( time ( 0 )-start ) << " seconds" << std::endl;
  std::cout << "Number of calls to getFirstIntersection() : " << sc->getCounter() << std::endl;
}

basic_framebuffer::basic_framebuffer ( size_t w_,size_t h_ , Gscene *sc_ ) :w ( w_ ),h ( h_ ),mem ( w_*h_ ),framebuffer ( sc_ )
{
  if (mem)
    buffer=new Color[mem];
  else
    buffer = NULL;
}

void basic_framebuffer::resize ( size_t w_,size_t h_ )
{
  size_t mem2=w_*h_;

  if ( mem2>mem )
  {
    delete buffer;
    mem=mem2;
    buffer=new Color[mem];
  }

  w=w_;
  h=h_;
  empty=true;
}

void basic_framebuffer::clear ( void )
{
  empty=true;
}

void basic_framebuffer::save_image ( std::ostream& file )
{
  file.write ( ( char* ) &w,sizeof ( w ) );
  file.write ( ( char* ) &h,sizeof ( h ) );
  file.write ( ( char* ) buffer,sizeof ( Color ) *w*h );
}
// kate: indent-mode cstyle; indent-width 2; replace-tabs on; 
